///////////////////////////////////////////////////////////////////////////////
//
//  (C) Copyright 1995 - 2005 OSR Open Systems Resources, Inc.
//  All Rights Reserved
//
//  This software is part of a licensed software product and may
//  only be used or copy in accordance with the terms of that license.
//
//  OSR Open Systems Resources, Inc.
//  105 Route 101A Suite 19
//  Amherst, NH 03031  (603) 595-6500 FAX: (603) 595-6503
//  email bugs to: bugs@osr.com
//
//
//  MODULE:
//
//    $Workfile: fileid.cpp $
//
//  ABSTRACT:
//
//    This code is part of the OSR bitbucket file system example; it is specifically concerned
//    with implementing both "open by file ID" as well as "open by object ID".  It demonstrates
//    a basic model for an in-line hash table implementation.   While this code is sufficiently
//    "stand alone" to demonstrate hashing, it is actually part of a larger (working) example
//    that supports a file system supporting open by file ID and open by object id.
//
//  AUTHOR:
//
//    OSR Open Systems Resources, Inc.
// 
//  REVISION:   
//
//    $Revision: 3 $
//
//
///////////////////////////////////////////////////////////////////////////////


//
// This contains definitions of the common driver data structures.
//
//
// For the purposes of describing the hashing logic the interesting data struture is our
// definition of the ID_TABLE_BUCKET:
//
//
// typedef struct _BB_ID_TABLE_BUCKET {
//     LIST_ENTRY BucketList;
//     ERESOURCE  BucketLock;
// } BB_ID_TABLE_BUCKET, *PBB_ID_TABLE_BUCKET;
//
//
// And the (fixed) size of the table.  We chose this value because it's a "reasonable" size
// for our purposes - it works for a few thousand entries without too much overhead.  Were the size
// of the table expected to be in the 100k entry range, it would make sense to increase this size (and
// not use a UCHAR in the hash computations.)
//
// Power-of-two values do not provide field properties for the hashing but they are extremely efficient to compute
// in binary computer systems; the gain (from our perspective) is generally not worth the complexity to use prime
// numbers (we've done it before - it does work - but it can turn out to be rather expensive computationally).
//
// #define ID_TABLE_SIZE (256)
//
//
// Note that we do not have a lock across the entire table - this is because we never actually enumerate the entire table.  If
// we DID need to enumerate the table, we'd need a "whole table lock"  This lock would normally be acquired shared (for any bucket
// level access) but then would be acquired exclusive (for whole table enumeration).
//

#include "bitbucket.h"


//
// InitIdTables
//
//   Initializes the ID tables for the given volume
//
// Input:
//   VolumeNode - this is the volume for which we are initializing the table.
//
// Outputs:
//   None.
//
// Returns:
//   SUCCESS - done successfully
//   INSUFFICIENT_RESOURCES - allocation failure(s)
//
// Notes:
//   None.
//
NTSTATUS InitIdTables(PBB_VOLUME_NODE VolumeNode)
{
    ULONG index;

    //
    // In general, the file system stores most of its data structures in paged pool.  The table buckets, however,
    // incorporate a dispatcher object (the ERESOURCE) that must be stored in non-paged pool.  Thus these structures are
    // allocated in non-paged pool.
    //
    VolumeNode->FileIdTable = (PBB_ID_TABLE_BUCKET) ExAllocatePoolWithTag(NonPagedPool, ID_TABLE_SIZE * sizeof(BB_ID_TABLE_BUCKET) , 'iFbB');

    if (NULL == VolumeNode->FileIdTable) {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    //
    // There is one table for file IDs (above) and a second for object IDs (this line)
    //
    VolumeNode->ObjectIdTable = (PBB_ID_TABLE_BUCKET) ExAllocatePoolWithTag(NonPagedPool, ID_TABLE_SIZE * sizeof(BB_ID_TABLE_BUCKET), 'iObB');

    if (NULL == VolumeNode->ObjectIdTable) {
        ExFreePool(VolumeNode->FileIdTable);
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    //
    // Our allocation of file IDs - in this sample file system - is dynamic and we allocate storage information
    // for the file IDs (and the synchronization object we use to serialize access again.)
    //
    VolumeNode->VolumeFidInfo = (PBB_VOLUME_FID_INFO) ExAllocatePoolWithTag(NonPagedPool, sizeof(BB_VOLUME_FID_INFO), 'fVbB');

    if (NULL == VolumeNode->VolumeFidInfo) {
        ExFreePool(VolumeNode->FileIdTable);
        ExFreePool(VolumeNode->ObjectIdTable);
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    //
    // Having successfully allocated the buckets, we must initialize them.
    //
    for (index = 0; index < ID_TABLE_SIZE; index++) {

        InitializeListHead(&VolumeNode->FileIdTable[index].BucketList);
        ExInitializeResourceLite(&VolumeNode->FileIdTable[index].BucketLock);

        InitializeListHead(&VolumeNode->ObjectIdTable[index].BucketList);
        ExInitializeResourceLite(&VolumeNode->ObjectIdTable[index].BucketLock);
    }

    //
    // This sets up information we use for assigning the file IDs - this isn't really important in understanding hashing, but it
    // at least explains what we're trying to achieve here.
    //
    KeQuerySystemTime(&VolumeNode->VolumeFidInfo->NextFileId);
    VolumeNode->VolumeFidInfo->FileIdIncrement.HighPart = 0;
    VolumeNode->VolumeFidInfo->FileIdIncrement.LowPart = KeQueryTimeIncrement();
    KeInitializeSpinLock(&VolumeNode->VolumeFidInfo->FileIdSpinLock);

    return STATUS_SUCCESS;

}

//
// ShutdownIdTables
//
//   Tear down the ID tables built for the volume.
//
// Input:
//   VolumeNode - this is the volume for which we are tearing down the hash tables.
//
// Outputs:
//   None.
//
// Returns:
//   None - this is a VOID function
//
// Notes:
//   None.
//
VOID ShutdownIdTables(PBB_VOLUME_NODE VolumeNode)
{
    ULONG index;

    for (index = 0; index < ID_TABLE_SIZE; index++) {
        ExDeleteResourceLite(&VolumeNode->FileIdTable[index].BucketLock);
        ExDeleteResourceLite(&VolumeNode->ObjectIdTable[index].BucketLock);
    }

    ExFreePool(VolumeNode->FileIdTable);
    ExFreePool(VolumeNode->ObjectIdTable);
    VolumeNode->FileIdTable = 0;
    VolumeNode->ObjectIdTable = 0;

    return;
}

//
// HashFileId
//
//   Given a 64-bit file ID value, we compute an 8 bit hash value (note that this restricts
//   our table to 256 entries - we'd have to change this function in order to make the table bigger)
//
// Input:
//   FileId - this is the 64-bit file ID for which we are returning a hash value.
//
// Outputs:
//   None.
//
// Returns:
//   Computed hash value
//
// Notes:
//   This is implemented inline (presumably, there's not a lot of code here within an efficient compiler).  Note that
//   we compute the hash value by simply XOR-ing in the byte values of the file ID.  This assumes a reasonable "scatter" of
//   the file IDs (since our hash function should try to minimize collisions, which is best achieved by scattering the input set
//   uniformly across the hash values returned.)
//
__inline UCHAR HashFileId(ULONG64 FileId)
{
    PUCHAR fid = (PUCHAR) &FileId;
    UCHAR hash = 0;
    ULONG index;

    for (index = 0; index < sizeof(ULONG64); index++) {
        hash ^= fid[0];
    }

    return hash;
}

//
// HashObjectId
//
//   Given a 128-bit object ID value, we compute an 8 bit hash value (note that this restricts
//   our table to 256 entries - we'd have to change this function in order to make the table bigger)
//
// Input:
//   ObjectId - this is the 128-bit object ID for which we are returning a hash value.
//
// Outputs:
//   None.
//
// Returns:
//   Computed hash value
//
// Notes:
//   The object ID probably does have greater variability than that file ID, but the algorithm is
//   essentially the same between the two cases.
//   
//
__inline UCHAR HashObjectId(GUID *ObjectId)
{
    PUCHAR fid = (PUCHAR) ObjectId;
    UCHAR hash = 0;
    ULONG index;

    for (index = 0; index < sizeof(GUID); index++) {
        hash ^= fid[0];
    }

    return hash;
}

//
// FindNodeByFileId
//
//  This is our lookup function - we retrieve the entry (from the volume structure where we
//  stored it) and the object ID for which we are seeking.  Note that we lock the table
//
// Input:
//   FileId - this is the 64-bit file ID we're trying to find.  Note that we will return the
//            entry from the table if we actually do find it.
//
// Outputs:
//   None.
//
// Returns:
//   File (or directory) node that we find associated with the given file ID, or NULL if we don't find anything.
//
// Notes:
//   Locking is important here - note that because we are doing a lookup, we only use a shared lock on the table.  The
//   idea is to minimize (as much as possible) the contention on the hash table buckets.  This is one technique for permitting
//   smaller tables (lower contention) and helps in scaling.
//   
//
PBB_BASE_NODE FindNodeByFileId(PBB_VOLUME_NODE VolumeNode, ULONG64 FileId)
{
    UCHAR hash = HashFileId(FileId);
    PLIST_ENTRY le;
    PBB_BASE_NODE baseNode = NULL;

    KeEnterCriticalRegion();
    ExAcquireResourceSharedLite(&VolumeNode->FileIdTable[hash].BucketLock, TRUE);

    le = VolumeNode->FileIdTable[hash].BucketList.Flink;

    //
    // Within the bucket we're just looking through a doubly linked list.  Since we have the list
    // locked (shared) no other thread can insert or remove entries from the list.
    //
    while (le != &VolumeNode->FileIdTable[hash].BucketList) {
        baseNode = CONTAINING_RECORD(le, BB_BASE_NODE, FileIdListEntry);

        if (baseNode->FileId == FileId) {

            break;

        }

        baseNode = NULL;

        le = le->Flink;

    }

    ExReleaseResourceLite(&VolumeNode->FileIdTable[hash].BucketLock);
    KeLeaveCriticalRegion();

    return baseNode;
    
}

//
// FindNodeByObjectId
//
//  This is our lookup function - we retrieve the entry (from the volume structure where we
//  stored it) by using the object ID for which we are seeking.  Note that we lock the table
//
// Input:
//   ObjectId - this is the 128-bit object ID we're trying to find.  Note that we will return the
//              entry from the table if we actually do find it.
//
// Outputs:
//   None.
//
// Returns:
//   File (or directory) node that we find associated with the given object ID, or NULL if we don't find anything.
//
// Notes:
//   See the notes on FindNodeByFileId - the same comment applies
//   
//
PBB_BASE_NODE FindNodeByObjectId(PBB_VOLUME_NODE VolumeNode, GUID *Guid)
{
    UCHAR hash = HashObjectId(Guid);
    PLIST_ENTRY le;
    PBB_BASE_NODE baseNode = NULL;

    KeEnterCriticalRegion();
    ExAcquireResourceSharedLite(&VolumeNode->ObjectIdTable[hash].BucketLock, TRUE);

    le = VolumeNode->ObjectIdTable[hash].BucketList.Flink;

    while (le != &VolumeNode->ObjectIdTable[hash].BucketList) {
        baseNode = CONTAINING_RECORD(le, BB_BASE_NODE, ObjectIdListEntry);

        if (IsEqualGUID(baseNode->ObjectId, *Guid)) {

            break;

        }

        baseNode = NULL;

        le = le->Flink;
        
    }

    ExReleaseResourceLite(&VolumeNode->ObjectIdTable[hash].BucketLock);
    KeLeaveCriticalRegion();

    return baseNode;
    
}

//
// InsertNodeIdTables
//
//    This function is responsible for inserting a new file/directory node into the file and object ID tables.
//
// Input:
//    BaseNode - this points to the containing data structure (the "file or directory node" that we really use.)  Note that
//               we rely upon the existance of two LIST_ENTRY values in this data structure - one for each table into which we
//               are inserting the node.  We assume the node is not in the table.
//
// Outputs:
//   None.
//
// Returns:
//   VOID function.
//
// Notes:
//   This routine has no logic to protect against duplicate entries.  If there ARE duplicate entries, then when a subsequent lookup
//   is performed, the returned entry will be arbitrary.  The implementation we use for assigning these ID values does not permit
//   duplicate entries in any reasonable timeframe, so we don't protect against this possibility at this point in the code.
//   
//   Locking here is done EXCLUSIVE on the bucket.  It allows activity on other buckets, but for the duration of this operation no
//   other thread may access this bucket - thus, we keep the operation very short.
//
VOID InsertNodeIdTables(PBB_BASE_NODE BaseNode)
{
    UCHAR hash = HashFileId(BaseNode->FileId);
    PBB_VOLUME_NODE volumeNode = BaseNode->VolumeNode;

    KeEnterCriticalRegion();

    //
    // First the file ID
    //
    ExAcquireResourceExclusiveLite(&volumeNode->FileIdTable[hash].BucketLock, TRUE);
    InsertTailList(&volumeNode->FileIdTable[hash].BucketList, &BaseNode->FileIdListEntry);
    ExReleaseResourceLite(&volumeNode->FileIdTable[hash].BucketLock);


    //
    // Now the object ID
    //
    hash= HashObjectId(&BaseNode->ObjectId);
    
    ExAcquireResourceExclusiveLite(&volumeNode->ObjectIdTable[hash].BucketLock, TRUE);
    InsertTailList(&volumeNode->ObjectIdTable[hash].BucketList,&BaseNode->ObjectIdListEntry);
    ExReleaseResourceLite(&volumeNode->ObjectIdTable[hash].BucketLock);

    KeLeaveCriticalRegion();
}

//
// RemoveNodeIdTables
//
//    This function is responsible for removing an existing file/directory node from the file and object ID tables.
//
// Input:
//    BaseNode - this points to the containing data structure (the "file or directory node" that we really use.)  Note that
//               we rely upon the existance of two LIST_ENTRY values in this data structure - one for each table into which we
//               are inserting the node.  We assume the node is already in the table.
//
// Outputs:
//   None.
//
// Returns:
//   VOID function.
//
// Notes:
//   This routine has no logic to protect against duplicate entries.  If there ARE duplicate entries, then when a subsequent lookup
//   is performed, the returned entry will be arbitrary.  The implementation we use for assigning these ID values does not permit
//   duplicate entries in any reasonable timeframe, so we don't protect against this possibility at this point in the code.
//   
//
VOID RemoveNodeIdTables(PBB_BASE_NODE BaseNode)
{
    UCHAR hash = HashFileId(BaseNode->FileId);
    PBB_VOLUME_NODE VolumeNode = BaseNode->VolumeNode;

    KeEnterCriticalRegion();

    //
    // First the file ID
    //
    ExAcquireResourceExclusiveLite(&VolumeNode->FileIdTable[hash].BucketLock, TRUE);
    RemoveEntryList(&BaseNode->FileIdListEntry);
    ExReleaseResourceLite(&VolumeNode->FileIdTable[hash].BucketLock);

    //
    // Now the object ID
    //
    ExAcquireResourceExclusiveLite(&VolumeNode->ObjectIdTable[hash].BucketLock, TRUE);
    RemoveEntryList(&BaseNode->ObjectIdListEntry);
    ExReleaseResourceLite(&VolumeNode->ObjectIdTable[hash].BucketLock);
    
    KeLeaveCriticalRegion();

    //
    // Note that this is a bit of paranoid programming that can be removed when we're
    // confident we'll never double-remove the entries from the list.
    //
    InitializeListHead(&BaseNode->FileIdListEntry);
    InitializeListHead(&BaseNode->ObjectIdListEntry);
    
}




